from datetime import datetime

import tensorflow as tf
from imbDRL.data import get_train_test_val, load_imdb
from imbDRL.train.bandit import TrainCustomBandit
from imbDRL.utils import get_reward_distribution, rounded_dict
from tf_agents.bandits.environments.classification_environment import \
    ClassificationBanditEnvironment

training_loops = 1_000  # Total training loops
batch_size = 64  # Batch size of each step
steps_per_loop = 64  # Number of steps to take in the environment for each loop

conv_layers = None  # Convolutional layers
dense_layers = (256, 256)  # Dense layers
dropout_layers = (0.0, 0.3)  # Dropout layers

lr = 0.001  # Learning rate
min_epsilon = 0.01  # Minimal and final chance of choosing random action
decay_steps = 250  # Number of steps to decay from 1.0 to `min_epsilon`

model_dir = "./models/" + (NOW := datetime.now().strftime("%Y%m%d_%H%M%S"))
log_dir = "./logs/" + NOW

imb_rate = 0.01  # Imbalance rate
min_class = [0]  # Minority classes
maj_class = [1]  # Majority classes
X_train, y_train, X_test, y_test, = load_imdb()
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, imb_rate, min_class, maj_class)

reward_distr = get_reward_distribution(imb_rate)
train_ds = tf.data.Dataset.from_tensor_slices((X_train, y_train))
train_env = ClassificationBanditEnvironment(train_ds, reward_distr, batch_size)

model = TrainCustomBandit(training_loops, lr, min_epsilon, decay_steps, model_dir,
                          log_dir, batch_size=batch_size, steps_per_loop=steps_per_loop)

model.compile_model(train_env, conv_layers, dense_layers, dropout_layers)
model.train(X_val, y_val)
stats = model.evaluate(X_test, y_test)
print(rounded_dict(stats))
# ("Gmean", 0.28783) ("Fdot5", 0.109092) ("F1", 0.162604) ("F2", 0.319161) ("TP", 1114) ("TN", 1162) ("FP", 11338) ("FN", 136)



from imbDRL.data import get_train_test_val, load_imdb
from imbDRL.environments import ClassifyEnv
from imbDRL.examples.ddqn.example_classes import TrainCustomDDQN
from imbDRL.utils import rounded_dict
from tf_agents.environments.tf_py_environment import TFPyEnvironment
from tf_agents.utils import common

episodes = 50_000  # Total number of episodes
warmup_episodes = 10_000  # Amount of warmup steps to collect data with random policy
memory_length = warmup_episodes  # Max length of the Replay Memory

target_model_update = episodes // 50  # Period to overwrite the target Q-network with the default Q-network
target_update_tau = 1 / 50  # Soften the target model update

conv_layers = None  # Convolutional layers
dense_layers = (512, 256, )  # Dense layers
dropout_layers = (0.0, 0.2, )  # Dropout layers

lr = 0.005  # Learning rate
gamma = 0.0  # Discount factor
min_epsilon = 0.1  # Minimal and final chance of choosing random action
decay_episodes = episodes // 2  # Number of episodes to decay from 1.0 to `min_epsilon`
batch_size = 64

imb_rate = 0.1  # Imbalance rate
min_class = [0]  # Minority classes
maj_class = [1]  # Majority classes
X_train, y_train, X_test, y_test, = load_imdb()
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, imb_rate, min_class, maj_class)

# Change Python environment to TF environment
train_env = TFPyEnvironment(ClassifyEnv(X_train, y_train, imb_rate))

model = TrainCustomDDQN(episodes, warmup_episodes, lr, gamma, min_epsilon, decay_episodes,
                        target_model_update=target_model_update, target_update_tau=target_update_tau, batch_size=batch_size)

model.compile_model(train_env, conv_layers, dense_layers, dropout_layers, loss_fn=common.element_wise_huber_loss)
model.train(X_val, y_val)
stats = model.evaluate(X_test, y_test)
print(rounded_dict(stats))
# ('Gmean', 0.500371) ('Fdot5', 0.110592) ('F1', 0.15832) ('F2', 0.278524) ('TP', 705) ('TN', 5549) ('FP', 6951) ('FN', 545)



import tensorflow as tf
from imbDRL.data import get_train_test_val, load_creditcard
from imbDRL.examples.bandit.example_classes import TrainCustomBandit
from imbDRL.utils import get_reward_distribution, rounded_dict
from tf_agents.bandits.environments.classification_environment import \
    ClassificationBanditEnvironment

training_loops = 2_000  # Total training loops
batch_size = 32  # Batch size of each step
steps_per_loop = 32  # Number of steps to take in the environment for each loop

conv_layers = None   # Convolutional layers
dense_layers = (256, 256, )  # Dense layers
dropout_layers = (0.2, 0.2, )  # Dropout layers

lr = 0.001  # Learning Rate
min_epsilon = 0.01  # Minimal chance of choosing random action
decay_steps = training_loops // 3  # Number of episodes to decay from 1.0 to `min_epsilon`

imb_rate = 0.001729  # Imbalance rate
min_class = [1]  # Minority classes
maj_class = [0]  # Majority classes
X_train, y_train, X_test, y_test, = load_creditcard(normalization=True)
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, min_class, maj_class, val_frac=0.2)

reward_distr = get_reward_distribution(imb_rate)
train_ds = tf.data.Dataset.from_tensor_slices((X_train, y_train))
train_env = ClassificationBanditEnvironment(train_ds, reward_distr, batch_size)
model = TrainCustomBandit(training_loops, lr, min_epsilon, decay_steps, batch_size=batch_size, steps_per_loop=steps_per_loop)

model.compile_model(train_env, conv_layers, dense_layers, dropout_layers)
model.train(X_val, y_val)
stats = model.evaluate(X_test, y_test, X_train, y_train)
print(rounded_dict(stats))
# ("Gmean", 0.930651) ("Fdot5", 0.557743) ("F1", 0.643939) ("F2", 0.761649) ("TP", 85) ("TN", 56783) ("FP",  81) ("FN", 13)
# ("Gmean", 0.93917 ) ("Fdot5", 0.227749) ("F1", 0.315789) ("F2", 0.514793) ("TP", 87) ("TN", 56498) ("FP", 366) ("FN", 11)
# ("Gmean", 0.950929) ("Fdot5", 0.311189) ("F1", 0.412993) ("F2", 0.613793) ("TP", 89) ("TN", 56620) ("FP", 244) ("FN",  9)



import matplotlib.pyplot as plt
import numpy as np
from imbDRL.data import get_train_test_val, load_creditcard
from imbDRL.metrics import classification_metrics
from imbDRL.utils import rounded_dict
from tensorflow import keras
from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense, Dropout

imb_rate = 0.00173  # Imbalance rate
min_class = [1]  # Minority classes, same setup as in original paper
maj_class = [0]  # Majority classes
X_train, y_train, X_test, y_test, = load_creditcard(normalization=True)
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, min_class, maj_class, val_frac=0.2)

model = Sequential([Dense(256, activation="relu", input_shape=(X_train.shape[-1],)),
                    Dropout(0.2),
                    Dense(256, activation="relu"),
                    Dropout(0.2),
                    Dense(1, activation="sigmoid")])

model.summary()
metrics = [keras.metrics.FalseNegatives(name="fn"),
           keras.metrics.FalsePositives(name="fp"),
           keras.metrics.Precision(name="precision"),
           keras.metrics.Recall(name="recall")]
model.compile(optimizer=keras.optimizers.Adam(0.001), loss="binary_crossentropy", metrics=metrics)

model.fit(X_train,
          y_train,
          epochs=30,
          batch_size=2048,
          validation_data=(X_val, y_val))

y_pred = model(X_test).numpy()
stats = classification_metrics(y_test, np.around(y_pred).astype(int))
print(rounded_dict(stats))

f1s = []
thresholds = np.arange(0, 1, 0.01)
for th in thresholds:
    y_labels = (y_pred >= th).astype(int)
    stats = classification_metrics(y_test, y_labels).get("F1")
    f1s.append(stats)
ix = np.argmax(f1s)

print('Best Threshold=%f, F-Score=%.3f' % (thresholds[ix], f1s[ix]))
stats = classification_metrics(y_test, (y_pred >= thresholds[ix]).astype(int))
print(rounded_dict(stats))
plt.plot(thresholds, f1s)
plt.show()
# {'Gmean': 0.892088, 'F1': 0.852459, 'Sensitivity': 0.795918, 'Specificity': 0.999877, 'TP': 78, 'TN': 56857, 'FP': 7, 'FN': 20}
# {'Gmean': 0.920204, 'F1': 0.864583, 'Sensitivity': 0.846939, 'Specificity': 0.999807, 'TP': 83, 'TN': 56853, 'FP': 11, 'FN': 15}



from imbDRL.examples.ddqn.example_classes import TrainCartPole
from tf_agents.environments import suite_gym
from tf_agents.environments.tf_py_environment import TFPyEnvironment

episodes = 20_000  # Total number of episodes
warmup_episodes = 1_000  # Amount of warmup steps to collect data with random policy
memory_length = 100_000  # Max length of the Replay Memory

conv_layers = None
dense_layers = (128, )  # For CartPole v0, only a small network is needed.
dropout_layers = None

lr = 0.001  # Learning rate
gamma = 0.99  # Discount factor
min_epsilon = 0.1  # Minimal and final chance of choosing random action
decay_episodes = 100  # Number of episodes to decay from 1.0 to `min_epsilon`

# Change OpenAI Gym environment to Python environment to TF environment
train_env = TFPyEnvironment(suite_gym.load("CartPole-v0"))
val_env = TFPyEnvironment(suite_gym.load("CartPole-v0"))

model = TrainCartPole(episodes, warmup_episodes, lr, gamma, min_epsilon, decay_episodes)
model.compile_model(train_env, conv_layers, dense_layers, dropout_layers)
model.train(val_env, 10)  # Validate for 10 episodes each time
model.evaluate(val_env, 10)  # TODO: add video as evaluation



from imbDRL.data import get_train_test_val, load_image
from imbDRL.environments import ClassifyEnv
from imbDRL.examples.ddqn.example_classes import TrainCustomDDQN
from imbDRL.utils import rounded_dict
from tf_agents.environments.tf_py_environment import TFPyEnvironment

episodes = 50_000  # Total episodes, 120_000 in original paper, the original code only trains every 4 steps
warmup_episodes = 50_000  # Amount of warmup steps to collect data with random policy
memory_length = 50_000  # Max length of the Replay Memory
collect_steps_per_episode = 1  # Since train_interval=4 in original code and `episodes` is divided by 4
target_model_update = 2_500  # Since train_interval=4 in original code and `episodes` is divided by 4, update target model 4x faster
target_update_tau = 0.2  # Soften the target model update

conv_layers = ((32, (5, 5), 2), (32, (5, 5), 2), )  # Convolutional layers
dense_layers = (256, )  # Dense layers
dropout_layers = None  # Dropout layers

lr = 0.001  # Learning rate
gamma = 0.0  # Discount factor
min_epsilon = 0.01  # Minimal and final chance of choosing random action
decay_episodes = 25_000  # Number of episodes to decay from 1.0 to `min_epsilon`, divided by 4

imb_rate = 0.01  # Imbalance rate
min_class = [2]  # Minority classes, same setup as in original paper
maj_class = [0, 1, 3, 4, 5, 6, 7, 8, 9]  # Majority classes
X_train, y_train, X_test, y_test, = load_image("mnist")
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(
    X_train, y_train, X_test, y_test, min_class, maj_class, imb_rate=imb_rate)

# Change Python environment to TF environment
train_env = TFPyEnvironment(ClassifyEnv(X_train, y_train, imb_rate))

model = TrainCustomDDQN(episodes, warmup_episodes, lr, gamma, min_epsilon, decay_episodes, target_update_tau=target_update_tau,
                        collect_steps_per_episode=collect_steps_per_episode, target_model_update=target_model_update)

model.compile_model(train_env, conv_layers, dense_layers, dropout_layers)
model.train(X_val, y_val)
stats = model.evaluate(X_test, y_test, X_val, y_val)
print(rounded_dict(stats))
# ("Gmean", 0.991757) ("Fdot5", 0.699523) ("F1", 0.785714) ("F2", 0.89613) ("TP", 88) ("TN", 8921) ("FP", 47) ("FN", 1)



import pickle

import numpy as np
import tensorflow as tf
from imbDRL.metrics import (classification_metrics, network_predictions,
                            plot_pr_curve, plot_roc_curve)
from imbDRL.train.ddqn import TrainDDQN
from tf_agents.policies.policy_saver import PolicySaver


class TrainCustomDDQN(TrainDDQN):
    """Class for the custom training environment."""

    def collect_metrics(self, X_val, y_val):
        """Collects metrics using the trained Q-network."""
        y_pred = network_predictions(self.agent._target_q_network, X_val)
        stats = classification_metrics(y_val, y_pred)
        avgQ = np.mean(np.max(self.agent._target_q_network(X_val)[0].numpy(), axis=1))  # Max action for each x in X

        if not hasattr(self, 'best_f1'):  # If no best model yet
            self.best_f1 = 0.0

        if stats.get("F1") >= self.best_f1:  # Overwrite best model
            self.save_model()  # Saving directly to avoid shallow copy without trained weights
            self.best_f1 = stats.get("F1")

        with self.writer.as_default():
            tf.summary.scalar("AverageQ", avgQ, step=self.global_episode)

            for k, v in stats.items():
                tf.summary.scalar(k, v, step=self.global_episode)

    def evaluate(self, X_test, y_test, X_train=None, y_train=None):
        """
        Final evaluation of trained Q-network with X_test and y_test.
        Optional PR and ROC curve comparison to X_train, y_train to ensure no overfitting is taking place.
        """
        best_model = self.load_model(self.model_dir)

        if (X_train is not None) and (y_train is not None):
            plot_pr_curve(best_model, X_test, y_test, X_train, y_train)
            plot_roc_curve(best_model, X_test, y_test, X_train, y_train)

        y_pred = network_predictions(best_model, X_test)
        return classification_metrics(y_test, y_pred)

    def save_model(self):
        """Saves Q-network as pickle to `model_dir`."""
        with open(self.model_dir + ".pkl", "wb") as f:  # Save Q-network as pickle
            pickle.dump(self.agent._target_q_network, f)

    @staticmethod
    def load_model(fp: str):
        """Static method to load Q-network pickle from given filepath."""
        with open(fp + ".pkl", "rb") as f:  # Load the Q-network
            network = pickle.load(f)
        return network


class TrainCartPole(TrainDDQN):
    """Class for the CartPole environment."""

    def collect_metrics(self, val_env, val_episodes: int):
        """Calculates the average return for `val_episodes` using the trained policy."""
        total_return = 0.0
        for _ in range(val_episodes):
            time_step = val_env.reset()
            episode_return = 0.0

            while not time_step.is_last():
                action_step = self.agent.policy.action(time_step)
                time_step = val_env.step(action_step.action)
                episode_return += time_step.reward
            total_return += episode_return

        avg_return = total_return // val_episodes

        with self.writer.as_default():
            tf.summary.scalar("avg_return", avg_return.numpy()[0], step=self.global_episode)

    def evaluate(self, test_env, test_episodes: int):
        """Final evaluation of policy."""
        return self.collect_metrics(test_env, test_episodes)

    def save_model(self):
        """Saves the policy to `model_dir`."""
        saver = PolicySaver(self.agent.policy)
        saver.save(self.model_dir)

    @staticmethod
    def load_model(fp: str):
        """Loads a saved policy from given filepath."""
        return tf.saved_model.load(fp)



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import tensorflow_probability as tfp
from sklearn.model_selection import train_test_split


def plot_confusion_matrix(TP: int, FN: int, FP: int, TN: int) -> None:  # pragma: no cover
    """Plots confusion matric of given TP, FN, FP, TN.

    :param TP: True Positive
    :type  TP: int
    :param FN: False Negative
    :type  FN: int
    :param FP: False Positive
    :type  FP: int
    :param TN: True Negative
    :type  TN: int

    :return: None
    :rtype: NoneType
    """
    if not all(isinstance(i, int) for i in (TP, FN, FP, TN)):
        raise ValueError("Not all arguments are integers.")

    ticklabels = ("Minority", "Majority")
    sns.heatmap(((TP, FN), (FP, TN)), annot=True, fmt="_d", cmap="viridis", xticklabels=ticklabels, yticklabels=ticklabels)

    plt.title("Confusion matrix")
    plt.xlabel("Predicted labels")
    plt.ylabel("True labels")
    plt.show()


def split_csv(fp: str = "./data/creditcard.csv", fp_dest: str = "./data",
              name: str = "credit", test_size: int = 0.5, strat_col: str = "Class") -> None:
    """Splits a csv file in two, in a stratified fashion.
    Format for filenames will be `{name}0.csv and `{name}1.csv`.

    :param fp: The path at which the csv file is located.
    :type  fp: str
    :param fp_dest: The path to save the train and test files.
    :type  fp_dest: str
    :param name: The prefix for the files.
    :type  name: str
    :param test_size: The fraction of total size for the test file.
    :type  test_size: float
    :param strat_col: The column in the original csv file to stratify.

    :return: None, two files located at `fp_dest`.
    :rtype: NoneType
    """
    if not os.path.isfile(fp):
        raise FileNotFoundError(f"File at {fp=} does not exist.")
    if not os.path.isdir(fp_dest):
        raise ValueError(f"Directory at {fp_dest=} does not exist.")
    if not 0 < test_size < 1:
        raise ValueError(f"{test_size} is not in interval 0 < x < 1.")

    df = pd.read_csv(fp)

    if not (strat_col in df.columns):
        raise ValueError(f"Stratify column {strat_col} not found in DataFrame.")

    train, test = train_test_split(df, test_size=test_size, stratify=df[strat_col])

    train.to_csv(f"{fp_dest}/{name}0.csv", index=False)
    test.to_csv(f"{fp_dest}/{name}1.csv", index=False)


def get_reward_distribution(imb_rate: float):
    """Generates reward distribution based on given imbalance rate.

    From the documentation of tf_agents.bandits.environments.ClassificationBanditEnvironment:

    reward_distribution: a `tfd.Distribution` with event_shape
        `[num_classes, num_actions]`. Entry `[i, j]` is the reward for taking
        action `j` for an instance of class `i`.

    The environment will always have 2 classes with each 2 actions.
    Array[0][0] corresponds to action 0 on class 0, thus the majority class, thus True Negative
    Array[0][1]: False Positive reward
    Array[1][0]: False Negative reward
    Array[1][1]: True Positive reward

    :param imb_rate: The imbalance rate of the data: len(min_class) / len(maj_class)
    :type  imb_rate: float

    :return: Combined Independent distribution
    :rtype: tensorflow_probability.Distribution
    """
    if not 0 < imb_rate < 1:
        raise ValueError(f"{imb_rate} is not in interval 0 < x < 1.")

    bernoulli = tfp.distributions.Bernoulli(probs=[[0, 0], [0, 0]], dtype=np.float32)
    combined = (tfp.bijectors.Shift([[imb_rate, -imb_rate], [-1, 1]])
                (tfp.bijectors.Scale([[1, 1], [1, 1]])
                 (bernoulli)))
    return tfp.distributions.Independent(combined, reinterpreted_batch_ndims=2)


def rounded_dict(d: dict, precision: int = 6) -> dict:
    """Rounds all values in a dictionairy to `precision` digits after the decimal point.

    :param d: Dictionairy containing only floats or ints as values
    :type  d: dict

    :return: Rounded dictionairy
    :rtype: dict
    """
    return {k: round(v, precision) for k, v in d.items()}



from imbDRL.data import get_train_test_val, load_creditcard
from imbDRL.metrics import classification_metrics
from imbDRL.utils import rounded_dict
from xgboost import XGBClassifier

imb_rate = 0.00173  # Imbalance rate
min_class = [1]  # Minority classes, same setup as in original paper
maj_class = [0]  # Majority classes
X_train, y_train, X_test, y_test, = load_creditcard(normalization=True)
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, imb_rate, min_class, maj_class)
scale = (y_train.shape[0] - y_train.sum()) // y_train.sum()  # Proportion of majority to minority rows: ~578

model = XGBClassifier(objective="binary:logitraw", scale_pos_weight=scale, eval_metric="aucpr", max_delta_step=1)
model.fit(X_train, y_train)
y_pred = model.predict(X_test)

stats = classification_metrics(y_test, y_pred)
print(rounded_dict(stats))
# ("Gmean", 0.92026 ) ("Fdot5", 0.930493) ("F1", 0.897297) ("F2", 0.866388) ("TP", 83) ("TN", 56860) ("FP", 4) ("FN", 15)



from abc import ABC, abstractmethod
from datetime import datetime

import numpy as np
import tensorflow as tf
from tensorflow.keras.optimizers import Adam
from tf_agents.bandits.agents.examples.v2.trainer import get_training_loop_fn
from tf_agents.bandits.agents.neural_epsilon_greedy_agent import \
    NeuralEpsilonGreedyAgent
from tf_agents.drivers.dynamic_step_driver import DynamicStepDriver
from tf_agents.networks.q_network import QNetwork
from tf_agents.replay_buffers.tf_uniform_replay_buffer import \
    TFUniformReplayBuffer
from tqdm import tqdm


class TrainBandit(ABC):
    """Wrapper for DDQN training, validation, saving etc."""

    def __init__(self, episodes: int, lr: float, min_epsilon: float, decay_episodes: int, model_dir: str = None, log_dir: str = None,
                 batch_size: int = 64, steps_per_loop: int = 64, val_every: int = None) -> None:
        """
        Wrapper to make training easier.
        Code is partly based of https://www.tensorflow.org/agents/tutorials/bandits_tutorial

        :param episodes: Number of training episodes
        :type  episodes: int
        :param lr: Learning Rate for the Adam Optimizer
        :type  lr: float
        :param min_epsilon: Lowest and final value for epsilon
        :type  min_epsilon: float
        :param decay_episodes: Amount of episodes to decay from 1 to `min_epsilon`
        :type  decay_episodes: int
        :param model_dir: Location to save the trained models
        :type  model_dir: str
        :param log_dir: Location to save the logs, usefull for TensorBoard
        :type  log_dir: str
        :param batch_size: Number of samples in minibatch to train on each step
        :type  batch_size: int
        :param steps_per_loop: Number of steps to take each episode
        :type  steps_per_loop: int
        :param val_every: Validate the model every X episodes using the `collect_metrics()` function
        :type  val_every: int

        :return: None
        :rtype: NoneType
        """
        self.episodes = episodes  # Total episodes
        self.batch_size = batch_size  # Batch size of Replay Memory
        self.steps_per_loop = steps_per_loop

        if val_every is not None:
            self.val_every = val_every  # Validate the policy every `VAL_EVERY` episodes
        else:
            self.val_every = episodes // min(50, self.episodes)  # Can't validate the model 50 times if self.episodes < 50

        self.lr = lr  # Learning Rate
        self.min_epsilon = min_epsilon  # Minimal chance of choosing random action
        self.decay_episodes = decay_episodes  # Number of episodes to decay from 1.0 to `EPSILON`

        NOW = datetime.now().strftime("%Y%m%d_%H%M%S")
        if model_dir is None:
            self.model_dir = "./models/" + NOW
        else:
            self.model_dir = model_dir

        if log_dir is None:
            self.log_dir = "./logs/" + NOW
        else:
            self.log_dir = log_dir

        self.writer = tf.summary.create_file_writer(self.log_dir)
        self.global_episode = tf.Variable(0, name="global_episode", dtype=np.int64, trainable=False)  # Global train episode counter

        # Custom epsilon decay: https://github.com/tensorflow/agents/issues/339
        self.epsilon_decay = tf.compat.v1.train.polynomial_decay(
            1.0, self.global_episode, self.decay_episodes, end_learning_rate=self.min_epsilon)
        self.optimizer = Adam(learning_rate=self.lr)
        self.compiled = False

    def compile_model(self, train_env, conv_layers: tuple, dense_layers: tuple, dropout_layers: tuple,
                      loss_fn=tf.compat.v1.losses.mean_squared_error) -> None:
        """Initializes the Q-network, agent, collect policy and replay buffer.

        :param train_env: The training environment used by the agent
        :type  train_env: tf_agents.environments.tf_py_environment.TFPyEnvironment
        :param conv_layers: Tuple of architecture of the convolutional layers.
            From tf_agents.networks.q_network:
                (...) where each item is a length-three tuple indicating (filters, kernel_size, stride).
        :type  conv_layers: tuple
        :param dense_layers: Tuple of dense layer architecture. Each number in the tuple is the number of neurons of that layer.
        :type  dense_layers: tuple
        :param dropout_layers: Tuple of percentage of dropout per each dense layer
        :type  dropout_layers: tuple
        :param loss_fn: Callable loss function
        :type  loss_fn: tf.compat.v1.losses

        :return: None
        :rtype: NoneType
        """
        for layer in (conv_layers, dense_layers, dropout_layers):
            if not isinstance(layer, (tuple, list, type(None))):
                raise TypeError(f"Layer {layer=} must be tuple or None, not {type(layer)}.")

        self.train_env = train_env
        self.conv_layers = conv_layers
        self.dense_layers = dense_layers
        self.dropout_layers = dropout_layers
        self.loss_fn = loss_fn

        self.q_net = QNetwork(self.train_env.observation_spec(),
                              self.train_env.action_spec(),
                              conv_layer_params=self.conv_layers,
                              fc_layer_params=self.dense_layers,
                              dropout_layer_params=self.dropout_layers)

        self.agent = NeuralEpsilonGreedyAgent(train_env.time_step_spec(),
                                              train_env.action_spec(),
                                              reward_network=self.q_net,
                                              optimizer=Adam(learning_rate=self.lr),
                                              epsilon=self.epsilon_decay,
                                              train_step_counter=self.global_episode,
                                              error_loss_fn=loss_fn)

        self.replay_buffer = TFUniformReplayBuffer(self.agent.policy.trajectory_spec,
                                                   self.train_env.batch_size,
                                                   self.steps_per_loop)

        self.driver = DynamicStepDriver(env=self.train_env,
                                        policy=self.agent.collect_policy,
                                        num_steps=self.steps_per_loop * self.train_env.batch_size,
                                        observers=[self.replay_buffer.add_batch])

        self.training_loop = get_training_loop_fn(self.driver, self.replay_buffer, self.agent, self.steps_per_loop)
        self.compiled = True

    def train(self, *args) -> None:
        """Starts the training of the model. Includes metrics collection and model saving.

        :param *args: All arguments will be passed to `collect_metrics()`.
            This can be usefull to pass callables, testing environments or validation data.
        :type  *args: Any

        :return: None
        :rtype: NoneType, last step is saving the model as a side-effect
        """
        assert self.compiled, "Model must be compiled with model.compile_model() before training."

        self.collect_metrics(*args)  # Initial collection for step 0
        for _ in tqdm(range(self.episodes)):
            loss_info = self.training_loop()

            if not self.global_episode % self.val_every:
                self.collect_metrics(*args)

                with self.writer.as_default():
                    tf.summary.scalar("train_loss", loss_info.loss, step=self.global_episode)

        self.save_model()

    @abstractmethod
    def collect_metrics(self):
        """*args given in train() will be passed to this function."""
        raise NotImplementedError

    @abstractmethod
    def evaluate(self):
        """Evaluation function to run after training with seperate train-dataset."""
        raise NotImplementedError

    @abstractmethod
    def save_model(self):
        """Abstract method for saving the model/network/policy to disk."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def load_model(fp: str):
        """Abstract method for loading the model/network/policy of disk."""
        raise NotImplementedError



from imbDRL.data import get_train_test_val, load_creditcard
from imbDRL.examples.ddqn.example_classes import TrainCustomDDQN
from imbDRL.metrics import (classification_metrics, network_predictions,
                            plot_pr_curve, plot_roc_curve)
from imbDRL.utils import rounded_dict

min_class = [1]  # Minority classes, same setup as in original paper
maj_class = [0]  # Majority classes
fp_model = "./models/20201110_125147"

X_train, y_train, X_test, y_test, = load_creditcard(normalization=True)
X_train, y_train, X_test, y_test, X_val, y_val = get_train_test_val(X_train, y_train, X_test, y_test, min_class, maj_class)

network = TrainCustomDDQN.load_model(fp_model)

y_pred_train = network_predictions(network, X_train)
y_pred_val = network_predictions(network, X_val)
y_pred_test = network_predictions(network, X_test)

stats = classification_metrics(y_train, y_pred_train)
print(f"Train: {rounded_dict(stats)}")
stats = classification_metrics(y_val, y_pred_val)
print(f"Val:   {rounded_dict(stats)}")
stats = classification_metrics(y_test, y_pred_test)
print(f"Test:  {rounded_dict(stats)}")

plot_pr_curve(network, X_test, y_test, X_train, y_train)
plot_roc_curve(network, X_test, y_test, X_train, y_train)



import os
from typing import Tuple

import numpy as np
from pandas import read_csv
from sklearn.model_selection import train_test_split
from tensorflow.keras.datasets import cifar10, fashion_mnist, imdb, mnist
from tensorflow.keras.preprocessing.sequence import pad_sequences

TrainTestData = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
TrainTestValData = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]


def load_image(data_source: str) -> TrainTestData:
    """
    Loads one of the following image datasets: {mnist, famnist, cifar10}.
    Normalizes the data. Returns X and y for both train and test datasets.
    Dtypes of X's and y's will be `float32` and `int32` to be compatible with `tf_agents`.

    :param data_source: Either mnist, famnist or cifar10
    :type  data_source: str

    :return: Tuple of (X_train, y_train, X_test, y_test) containing original split of train/test
    :rtype: tuple
    """
    reshape_shape = -1, 28, 28, 1

    if data_source == "mnist":
        (X_train, y_train), (X_test, y_test) = mnist.load_data()

    elif data_source == "famnist":
        (X_train, y_train), (X_test, y_test) = fashion_mnist.load_data()

    elif data_source == "cifar10":
        (X_train, y_train), (X_test, y_test) = cifar10.load_data()
        reshape_shape = -1, 32, 32, 3

    else:
        raise ValueError("No valid `data_source`.")

    X_train = X_train.reshape(reshape_shape).astype(np.float32)  # Float32 is the expected dtype for the observation spec in the env
    X_test = X_test.reshape(reshape_shape).astype(np.float32)

    X_train /= 255  # /= is not available when casting int to float: https://stackoverflow.com/a/48948461/10603874
    X_test /= 255

    y_train = y_train.reshape(y_train.shape[0], ).astype(np.int32)
    y_test = y_test.reshape(y_test.shape[0], ).astype(np.int32)

    return X_train, y_train, X_test, y_test


def load_imdb(config: Tuple[int, int] = (5_000, 500)) -> TrainTestData:
    """Loads the IMDB dataset. Returns X and y for both train and test datasets.

    :param config: Tuple of number of most frequent words and max length of each sequence.
    :type  config: str

    :return: Tuple of (X_train, y_train, X_test, y_test) containing original split of train/test
    :rtype: tuple
    """
    if not isinstance(config, (tuple, list)):
        raise TypeError(f"{type(config)} is no valid datatype for `config`.")
    if len(config) != 2:
        raise ValueError("Tuple length of `config` must be 2.")
    if not all(i > 0 for i in config):
        raise ValueError("All integers of `config` must be > 0.")

    (X_train, y_train), (X_test, y_test) = imdb.load_data(num_words=config[0])

    X_train = pad_sequences(X_train, maxlen=config[1])
    X_test = pad_sequences(X_test, maxlen=config[1])

    y_train = y_train.astype(np.int32)
    y_test = y_test.astype(np.int32)

    return X_train, y_train, X_test, y_test


def load_creditcard(fp_train: str = "./data/credit0.csv", fp_test: str = "./data/credit1.csv",
                    normalization: bool = False) -> TrainTestData:
    """
    Loads the Kaggle Credit Card Fraud dataset from local filepaths. Returns X and y for both train and test datasets.
    Option to normalize the data with min-max normalization.
    Source for dataset: https://www.kaggle.com/mlg-ulb/creditcardfraud

    :param fp_train: Location of the train csv-file
    :type  fp_train: str
    :param fp_test: Location of the test csv-file
    :type  fp_test: str
    :param normalization: Normalize the data with min-max normalization?
    :type  normalization: bool

    :return: Tuple of (X_train, y_train, X_test, y_test) containing original split of train/test
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    if not os.path.isfile(fp_train):
        raise FileNotFoundError(f"`fp_train` {fp_train} does not exist.")
    if not os.path.isfile(fp_test):
        raise FileNotFoundError(f"`fp_test` {fp_test} does not exist.")
    if not isinstance(normalization, bool):
        raise TypeError(f"`normalization` must be of type `bool`, not {type(normalization)}")

    X_train = read_csv(fp_train).astype(np.float32)  # DataFrames directly converted to float32
    X_test = read_csv(fp_test).astype(np.float32)

    y_train = X_train["Class"].astype(np.int32)  # 1: Fraud/Minority, 0: No fraud/Majority
    y_test = X_test["Class"].astype(np.int32)
    X_train.drop(columns=["Time", "Class"], inplace=True)  # Dropping `Time` since future data for the model could have another epoch
    X_test.drop(columns=["Time", "Class"], inplace=True)

    # Other data sources are already normalized. RGB values are always in range 0 to 255.
    if normalization:
        mini, maxi = X_train.min(axis=0), X_train.max(axis=0)
        X_train -= mini
        X_train /= maxi - mini
        X_test -= mini
        X_test /= maxi - mini

    return X_train.values, y_train.values, X_test.values, y_test.values  # Numpy arrays


def load_sepsis(fp_train: str = "./data/sepsis0.csv", fp_test: str = "./data/sepsis1.csv",
                normalization: bool = False) -> TrainTestData:
    """
    Loads the sepsis dataset from local filepaths. Returns X and y for both train and test datasets.
    Option to normalize the data with min-max normalization.
    Source for dataset: https://mimic-iv.mit.edu/

    :param fp_train: Location of the train csv-file
    :type  fp_train: str
    :param fp_test: Location of the test csv-file
    :type  fp_test: str
    :param normalization: Normalize the data with min-max normalization?
    :type  normalization: bool

    :return: Tuple of (X_train, y_train, X_test, y_test) containing original split of train/test
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    if not os.path.isfile(fp_train):
        raise FileNotFoundError(f"`fp_train` {fp_train} does not exist.")
    if not os.path.isfile(fp_test):
        raise FileNotFoundError(f"`fp_test` {fp_test} does not exist.")
    if not isinstance(normalization, bool):
        raise TypeError(f"`normalization` must be of type `bool`, not {type(normalization)}")

    X_train = read_csv(fp_train).astype(np.float32)  # DataFrames directly converted to float32
    X_test = read_csv(fp_test).astype(np.float32)

    y_train = X_train["sepsis"].astype(np.int32)  # 1: Sepsis during admission, 0: No sepsis during admission
    y_test = X_test["sepsis"].astype(np.int32)
    X_train.drop(columns=["subject_id", "hadm_id", "sepsis", "pco2"], inplace=True)  # Dropping Subject and admission columns
    X_test.drop(columns=["subject_id", "hadm_id", "sepsis", "pco2"], inplace=True)  # Dropping pco2

    # Other data sources are already normalized. RGB values are always in range 0 to 255.
    if normalization:
        mini, maxi = X_train.min(axis=0), X_train.max(axis=0)
        X_train -= mini
        X_train /= maxi - mini
        X_test -= mini
        X_test /= maxi - mini

    return X_train.values, y_train.values, X_test.values, y_test.values  # Numpy arrays


def load_aki(fp_train: str = "./data/aki0.csv", fp_test: str = "./data/aki1.csv",
             normalization: bool = False) -> TrainTestData:
    """
    Loads the AKI dataset from local filepaths. Returns X and y for both train and test datasets.
    Option to normalize the data with min-max normalization.
    Source for dataset: https://mimic-iv.mit.edu/

    :param fp_train: Location of the train csv-file
    :type  fp_train: str
    :param fp_test: Location of the test csv-file
    :type  fp_test: str
    :param normalization: Normalize the data with min-max normalization?
    :type  normalization: bool

    :return: Tuple of (X_train, y_train, X_test, y_test) containing original split of train/test
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    if not os.path.isfile(fp_train):
        raise FileNotFoundError(f"`fp_train` {fp_train} does not exist.")
    if not os.path.isfile(fp_test):
        raise FileNotFoundError(f"`fp_test` {fp_test} does not exist.")
    if not isinstance(normalization, bool):
        raise TypeError(f"`normalization` must be of type `bool`, not {type(normalization)}")

    X_train = read_csv(fp_train).astype(np.float32)  # DataFrames directly converted to float32
    X_test = read_csv(fp_test).astype(np.float32)

    y_train = X_train["aki"].astype(np.int32)  # 1: Sepsis during admission, 0: No sepsis during admission
    y_test = X_test["aki"].astype(np.int32)
    X_train.drop(columns=["hadm_id", "aki"], inplace=True)  # Dropping Subject and admission columns
    X_test.drop(columns=["hadm_id", "aki"], inplace=True)  # Dropping pco2

    # Other data sources are already normalized. RGB values are always in range 0 to 255.
    if normalization:
        mini, maxi = X_train.min(axis=0), X_train.max(axis=0)
        X_train -= mini
        X_train /= maxi - mini
        X_test -= mini
        X_test /= maxi - mini

    return X_train.values, y_train.values, X_test.values, y_test.values  # Numpy arrays


def get_train_test_val(X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray, y_test: np.ndarray, min_classes: list,
                       maj_classes: list, imb_rate: float = None, val_frac: float = 0.25, print_stats: bool = True) -> TrainTestValData:
    """
    Imbalances data and divides the data into train, test and validation sets.
    The imbalance rate of each individual dataset is approx. the same as the given `imb_rate`.

    :param X_train: The X_train data
    :type  X_train: np.ndarray
    :param y_train: The y_train data
    :type  y_train: np.ndarray
    :param X_test: The X_test data
    :type  X_test: np.ndarray
    :param y_test: The y_test data
    :type  y_test: np.ndarray
    :param min_classes: List of labels of all minority classes
    :type  min_classes: list
    :param maj_classes: List of labels of all majority classes.
    :type  maj_classes: list
    :param imb_rate: Imbalance ratio for minority to majority class: len(minority datapoints) / len(majority datapoints)
        If the `imb_rate` is None, data will not be imbalanced and will only be relabeled to 1's and 0's.
    :type  imb_rate: float
    :param val_frac: Fraction to take from X_train and y_train for X_val and y_val
    :type  val_frac: float
    :param print_stats: Print the imbalance ratio of the imbalanced data?
    :type  print_stats: bool

    :return: Tuple of (X_train, y_train, X_test, y_test, X_val, y_val)
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    if not 0 < val_frac < 1:
        raise ValueError(f"{val_frac} is not in interval 0 < x < 1.")
    if not isinstance(print_stats, bool):
        raise TypeError(f"`print_stats` must be of type `bool`, not {type(print_stats)}.")

    X_train, y_train = imbalance_data(X_train, y_train, min_classes, maj_classes, imb_rate=imb_rate)  # Imbalance the data
    X_test, y_test = imbalance_data(X_test, y_test, min_classes, maj_classes, imb_rate=imb_rate)

    # stratify=y_train to ensure class balance is kept between train and validation datasets
    X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=val_frac, stratify=y_train)

    if print_stats:
        p_train, p_test, p_val = [((y == 1).sum(), (y == 1).sum() / (y == 0).sum()) for y in (y_train, y_test, y_val)]
        print(f"Imbalance ratio `p`:\n"
              f"\ttrain:      n={p_train[0]}, p={p_train[1]:.6f}\n"
              f"\ttest:       n={p_test[0]}, p={p_test[1]:.6f}\n"
              f"\tvalidation: n={p_val[0]}, p={p_val[1]:.6f}")

    return X_train, y_train, X_test, y_test, X_val, y_val


def imbalance_data(X: np.ndarray, y: np.ndarray, min_class: list, maj_class: list, imb_rate: float = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Split data in minority and majority, only values in {min_class, maj_class} will be kept.
    (Possibly) decrease minority rows to match the imbalance rate.
    If initial imb_rate of dataset is lower than given `imb_rate`, the imb_rate will not be changed.
    If the `imb_rate` is None, data will not be imbalanced and will only be relabeled to 1's and 0's.

    Note: Data will not be shuffled
    """
    if not isinstance(X, np.ndarray):
        raise TypeError(f"`X` must be of type `np.ndarray` not {type(X)}")
    if not isinstance(y, np.ndarray):
        raise TypeError(f"`y` must be of type `np.ndarray` not {type(y)}")
    if X.shape[0] != y.shape[0]:
        raise ValueError("`X` and `y` must contain the same amount of rows.")
    if not isinstance(min_class, (list, tuple)):
        raise TypeError("`min_class` must be of type list or tuple.")
    if not isinstance(maj_class, (list, tuple)):
        raise TypeError("`maj_class` must be of type list or tuple.")

    if (imb_rate is not None) and not (0 < imb_rate < 1):
        raise ValueError(f"{imb_rate} is not in interval 0 < x < 1.")

    if imb_rate is None:  # Do not imbalance data if no `imb_rate` is given
        imb_rate = 1

    X_min, X_maj = [], []
    for i, value in enumerate(y):
        if value in min_class:
            X_min.append(X[i])

        if value in maj_class:
            X_maj.append(X[i])

    min_len = int((X_maj_len := len(X_maj)) * imb_rate)

    # Keep all majority rows, decrease minority rows to match `imb_rate`
    X_imb = np.array(X_maj + X_min[:min_len], dtype=np.float32)  # `min_len` could be more than the number of minority rows
    y_imb = np.concatenate((np.zeros(X_maj_len), np.ones(X_imb.shape[0] - X_maj_len))).astype(np.int32)

    return X_imb, y_imb



import pickle

import tensorflow as tf
from imbDRL.metrics import (classification_metrics, network_predictions,
                            plot_pr_curve, plot_roc_curve)
from imbDRL.train.bandit import TrainBandit


class TrainCustomBandit(TrainBandit):
    """Class for the bandit training environment."""

    def collect_metrics(self, X_val, y_val):
        """Collects metrics using the trained Q-network."""
        y_pred = network_predictions(self.agent._reward_network, X_val)
        stats = classification_metrics(y_val, y_pred)

        with self.writer.as_default():
            for k, v in stats.items():
                tf.summary.scalar(k, v, step=self.global_episode)

    def evaluate(self, X_test, y_test, X_train=None, y_train=None):
        """
        Final evaluation of trained Q-network with X_test and y_test.
        Optional PR and ROC curve comparison to X_train, y_train to ensure no overfitting is taking place.
        """
        if (X_train is not None) and (y_train is not None):
            plot_pr_curve(self.agent._reward_network, X_test, y_test, X_train, y_train)
            plot_roc_curve(self.agent._reward_network, X_test, y_test, X_train, y_train)

        y_pred = network_predictions(self.agent._reward_network, X_test)
        return classification_metrics(y_test, y_pred)

    def save_model(self):
        """Saves Q-network as pickle to `model_dir`."""
        with open(self.model_dir + ".pkl", "wb") as f:  # Save Q-network as pickle
            pickle.dump(self.agent._reward_network, f)

    @staticmethod
    def load_model(fp: str):
        """Static method to load Q-network pickle from given filepath."""
        with open(fp + ".pkl", "rb") as f:  # Load the Q-network
            network = pickle.load(f)
        return network
